#include "AWList.h"

#include <stdio.h>
#include <stdlib.h>

typedef enum
{
    TRUE  = 1,
    FALSE = 0
} BOOL;

// ================> SOURCE CODE BELOW <================
AWList* AWList_init(AWList* list)
{
    if (list == NULL)
    {
        list = calloc(1, sizeof(AWList));

        list->free_me = TRUE;
    }
    else
    {
        list->free_me = FALSE;
    }

    list->head = NULL;
    list->tail = NULL;
    list->curr = NULL;
    list->cnt  = 0;
    return list;
}

void AWList_destroy(AWList* list)
{
    if (list)
    {
        for (; list->head; AWList_delete_node(list, list->head))
            ;

        if (list->free_me == TRUE)
        {
            free(list);
            list = NULL;
        }
    }
}

void AWList_add_head(AWList* list, void* data, void (*delete_func)(void*))
{
    AWNode* node      = (AWNode*)malloc(sizeof(AWNode));
    node->data        = data;
    node->delete_func = delete_func;

    if (list->head == NULL)
    {
        node->next = NULL;

        list->tail = node;
    }
    else
    {
        node->next = list->head;
    }
    list->curr = list->head = node;
    list->cnt += 1;
}

void AWList_add_tail(AWList* list, void* data, void (*delete_func)(void*))
{
    AWNode* node      = (AWNode*)malloc(sizeof(AWNode));
    node->data        = data;
    node->next        = NULL;
    node->delete_func = delete_func;

    if (list->head == NULL)
    {
        list->curr = list->head = node;
    }
    else
    {
        list->tail->next = node;
    }
    list->tail = node;

    list->cnt += 1;
}

void AWList_delete_node(AWList* list, AWNode* node)
{
    if (list->head == NULL || node == NULL || node->delete_func == NULL) return;


    if (list->head == node)
    {
        list->curr = list->head = list->head->next;
    }
    else
    {
        AWNode* tmp = list->head;
        for (; tmp->next && tmp->next != node; tmp = tmp->next)
        {
        }
        if (tmp->next == node)
            tmp->next = node->next;
        else
            return;
    }

    node->delete_func(node->data);

    free(node);
    node = NULL;
    list->cnt -= 1;
}

void AWList_clear(AWList *list)
{
    int freeme = list->free_me;
    list->free_me = FALSE;

    AWList_destroy(list);

    list->free_me = freeme;
}

AWNode *AWList_find_node(AWList* list, void* data, int (*cmp_callback)(void*, void*))
{
    AWNode *node;
    AWList_for_each(node, list)
    {
        if (cmp_callback(node->data, data) == 0)
            return node;
    }

    return NULL;
}

AWNode* AWList_iterator(AWList* list)
{
    if (list->curr)
    {
        AWNode* curr = list->curr;
        list->curr = list->curr->next;
        return curr;
    }
    else
        return NULL;
}
